/******************************************************************************************

   TacTix.java - TacTix board game
   Version 2.0
   www.mazeworks.com
   
   Copyright (c) 2002 Robert Kirkland
   All Rights Reserved.

   Permission is hereby granted, free of charge, to any person obtaining a copy 
   of this software and associated documentation files (the "Software"), to
   deal in the Software without restriction, including without limitation the
   rights to use, copy, modify, merge, publish, distribute, and/or sell copies
   of the Software, and to permit persons to whom the Software is furnished to
   do so, provided that the above copyright notice and this permission notice
   appear in all copies of the Software and that both the above copyright
   notice and this permission notice appear in supporting documentation.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.
   IN NO EVENT SHALL THE COPYRIGHT HOLDER INCLUDED IN THIS NOTICE BE LIABLE FOR
   ANY CLAIM, OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Except as contained in this notice, the name of the copyright holder shall
   not be used in advertising or otherwise to promote the sale, use or other
   dealings in this Software without prior written authorization of the
   copyright holder.

******************************************************************************************/

import java.applet.* ;
import java.awt.* ;
import java.net.* ;
import java.util.* ;

/******** TACTIX ********
   main class: initializes applet panel, starts games,
   runs AI move thread, handles player moves and UI events
*/   
public class TacTix extends Applet implements Runnable {
   // applet info
   public String getAppletInfo() {
      return "TacTix 2.0 - Copyright (c) 2002 Robert Kirkland"; }
   String parameterInfo[][] = {
      {"bgColor",      "integer", "Background color (24-bit RGB hex int)"},
      {"title",        "string",  "Applet title text"},
      {"newGame",      "string",  "New Game button text"},
      {"move",         "string",  "Move button text"},
      {"beginner",     "string",  "Beginner level text"},
      {"intermediate", "string",  "Intermediate level text"},
      {"expert",       "string",  "Expert level text"},
      {"yourMove",     "string",  "Player Move status text"},
      {"thinking",     "string",  "Computer Thinking status text"},
      {"playerWin",    "string",  "Player Win status text"},
      {"tactixWin",    "string",  "Computer Win status text"} } ;
   public String[][] getParameterInfo() { return parameterInfo ; }

   static final Color  DF_BG_COLOR  =Color.white ;   // default background color
   static final String DF_TITLE     ="TACTIX",       // default text strings
                       DF_NEW_GAME  ="NEW GAME",
                       DF_MOVE      ="MOVE",
                       DF_BEGINNER  ="Beginner",  
                       DF_INTERMED  ="Intermediate",
                       DF_EXPERT    ="Expert",
                       DF_YOUR_MOVE ="Your move..",
                       DF_THINKING  ="Thinking..",
                       DF_PLAYER_WIN="You win !",
                       DF_TACTIX_WIN="Game over !" ;
   static final int bcWidth=214, bcHeight=214 ;
   protected Color BG_COLOR ;
   protected String TITLE, NEW_GAME, MOVE, BEGINNER, INTERMED, EXPERT, 
                    YOUR_MOVE, THINKING, PLAYER_WIN, TACTIX_WIN ;
	private boolean firstMove, gameOver ;
	private Board bd ;
	private BoardCanvas bc ;
   private Panel centerPanel, eastPanel, newgamePanel, movePanel,
              beginPanel, intermedPanel, expertPanel, bcPanel ;
	private Image boardImage, counterImage, counterOnImage ;
   private AI piet ;
   private ControlPanel cp ;
   private Thread pietMoveThread ;
	
   public void init() {
      // --- get applet parameters ---
      // background color
      String bgColorString = getParameter("bgColor") ;
      try { BG_COLOR = bgColorString==null ? DF_BG_COLOR : new Color(Integer.parseInt(bgColorString,16)) ; } 
      catch (NumberFormatException e) {
         BG_COLOR = DF_BG_COLOR ;
         System.out.println("ERROR: Invalid bgColor parameter.") ;
      }
      // text
      TITLE = getParameter("title") ; if (TITLE==null) TITLE = new String(DF_TITLE) ;
      NEW_GAME = getParameter("newGame") ; if (NEW_GAME==null) NEW_GAME = new String(DF_NEW_GAME) ;
      MOVE = getParameter("move") ; if (MOVE==null) MOVE = new String(DF_MOVE) ;
      BEGINNER = getParameter("beginner") ; if (BEGINNER==null) BEGINNER = new String(DF_BEGINNER) ;
      INTERMED = getParameter("intermediate") ; if (INTERMED==null) INTERMED = new String(DF_INTERMED) ;
      EXPERT = getParameter("expert") ; if (EXPERT==null) EXPERT = new String(DF_EXPERT) ;
      YOUR_MOVE = getParameter("yourMove") ; if (YOUR_MOVE==null) YOUR_MOVE = new String(DF_YOUR_MOVE) ;
      THINKING = getParameter("thinking") ; if (THINKING==null) THINKING = new String(DF_THINKING) ;
      PLAYER_WIN = getParameter("playerWin") ; if (PLAYER_WIN==null) PLAYER_WIN = new String(DF_PLAYER_WIN) ;
      TACTIX_WIN = getParameter("tactixWin") ; if (TACTIX_WIN==null) TACTIX_WIN = new String(DF_TACTIX_WIN) ;

		// load images
		MediaTracker tracker = new MediaTracker(this) ;
		URL url = getCodeBase() ;
		boardImage = getImage(url,"board.gif") ;
		counterImage = getImage(url,"counter.gif") ;
		counterOnImage = getImage(url,"counter1.gif") ;
		tracker.addImage(boardImage,0) ;
		tracker.addImage(counterImage,1) ; 
      tracker.addImage(counterOnImage,2) ;
		try { tracker.waitForAll(); }
		catch (InterruptedException e) {}

      // initialize applet layout
      setBackground(BG_COLOR) ;
      setLayout(new FlowLayout(FlowLayout.CENTER,20,10)) ;
      // create board canvas
      bc = new BoardCanvas(boardImage,counterImage,counterOnImage,this) ;
      bc.resize(bcWidth,bcHeight) ;
      // create control panel
      cp = new ControlPanel(this) ;
      Panel p1 = new Panel() ;
      p1.setBackground(Color.lightGray) ;
      p1.setLayout(new FlowLayout(FlowLayout.CENTER,5,5)) ;
      Panel p2 = new Panel() ;
      p2.setBackground(Color.gray) ;
      p2.setLayout(new FlowLayout(FlowLayout.CENTER,2,2)) ;
      p1.add(cp) ;
      p2.add(p1) ;
      add(bc) ;
      add(p2) ;
      validate() ;
      // force class file load
      Position p = new Position() ;
      Move m = new Move() ;
      BestMove bm = new BestMove() ;
      Ref r = new Ref() ;
		newGame() ;
	}
   // start game
   void newGame() {
      stop() ;
      System.gc() ;
      firstMove = true ;
      gameOver = false ; 
      bd = new Board(bc,this) ;
      piet = new AI(bd,cp,this) ;
      cp.setMoveButton(true) ;
      cp.setStatus(YOUR_MOVE) ;
      bc.drawBoard(bd) ;
      bc.requestFocus() ;
   }      
   // start AI move thread
   public void pietMoveStart() {
      if (pietMoveThread == null) {
         pietMoveThread = new Thread(this) ;
         pietMoveThread.start() ;
      }
   }
   // kill thread
   public void stop() {
      if (pietMoveThread != null) {
         pietMoveThread.stop() ;
         pietMoveThread = null ;
      }
   }
   // run thread
   public void run() { 
      cp.setStatus(THINKING) ;
      piet.move() ; 
      bc.drawBoard(bd) ;
      if (bd.getCount()==0) endGame(piet.PLAYER) ;
      else cp.setStatus(YOUR_MOVE) ;
      pietMoveThread = null ;
   }
   // board canvas mouse event
   void selectCounter(int i,int j) {
      cp.setMoveButton(true) ;
      bd.select(i,j) ;
      bc.drawBoard(bd) ;
   }
   // Move button event
   void moveButton() {
      if (!gameOver) {
         cp.setMoveButton(false) ;
         if (firstMove) {
            firstMove = false ;
            if (bd.getSelectCount()!=0) playerMove() ;
            pietMoveStart() ;
         } else if (bd.getSelectCount()!=0) {
            playerMove() ;
            if (bd.getCount()==0) endGame(piet.COMPUTER) ;
            else pietMoveStart() ;
         }
      }
      bc.requestFocus() ;
   }
   private void playerMove() {
      bd.makeMove() ;
      bc.drawBoard(bd) ;
   }
   private void endGame(int side) {
      if (side==piet.COMPUTER) cp.setStatus(TACTIX_WIN) ;
      else cp.setStatus(PLAYER_WIN) ;
      gameOver = true ;
   }
   void setFocus() { bc.requestFocus() ; }
   void sleep(int ms) {
      try { Thread.currentThread().sleep(ms) ; } 
      catch (InterruptedException e) {}
   }
}  
/******** BOARD ********
   controls board status, rules of the game
*/   
final class Board {
	static final int EMPTY=0, COUNTER=1, SELECTED=2 ;
   private BoardCanvas bc ;
   private TacTix main ;
	private int bd[][]=new int[4][4] ;
   
   // constructor
   Board(BoardCanvas bc,TacTix main) {
      this.bc = bc ;
      this.main = main ;
      for (int i=0; i<4; i++)
         for (int j=0; j<4; j++) bd[i][j] = COUNTER ;
   }
	boolean isEmpty(int i,int j) { return (bd[i][j] == EMPTY) ; }
   boolean isCounter(int i,int j) { return (bd[i][j] == COUNTER) ; }
   boolean isCounter(Point p) { return (bd[p.x][p.y] == COUNTER) ; }
   boolean isSelect(int i,int j) { return (bd[i][j] == SELECTED) ; }
   void setEmpty(int i,int j) { bd[i][j] = EMPTY ; return ; }
   void setEmpty(Point p) { bd[p.x][p.y] = EMPTY ; return ; }
   void setCounter(int i,int j) { bd[i][j] = COUNTER ; return ; }
   void setCounter(Point p) { bd[p.x][p.y] = COUNTER ; return ; }
   void setSelect(int i,int j) { bd[i][j] = SELECTED ; return ; }
   void setSelect(Point p) { bd[p.x][p.y] = SELECTED ; return ; }

   int[][] getBoard() { return bd ; }
   int getCount() {
      int count = 0 ;
      for (int i=0; i<4; i++)
         for (int j=0; j<4; j++) if (!isEmpty(i,j)) count++ ;
      return count ;            
   }
   int getSelectCount() {
      int count = 0 ;
      for (int i=0; i<4; i++)
         for (int j=0; j<4; j++) if (isSelect(i,j)) count++ ;
      return count ;            
   }
   private boolean isValidSelect(int col,int row) {
      if (!isCounter(col,row)||getSelectCount()>=4) return false ;
      for (int i=0; i<4; i++) {
         for (int j=0; j<4; j++) {
            if (isSelect(i,j)) {
               // check if adjacent in column
               if (col==i) {
                  if (row==0) {
                     if (!isSelect(col,row+1)) return false ;
                  }
                  else if (row==3) {
                     if (!isSelect(col,row-1)) return false ;
                  }
                  else
                     if ( (!isSelect(col,row+1))&
                          (!isSelect(col,row-1)) ) return false ;
               }
               // check if adjacent in row
               else if (row==j) {
                  if (col==0) {
                     if (!isSelect(col+1,row)) return false ;
                  }
                  else if (col==3) {
                     if (!isSelect(col-1,row)) return false ;
                  }
                  else
                     if ( (!isSelect(col+1,row))&
                          (!isSelect(col-1,row)) ) return false ;
               // not on same row or column
               } else return false ;
            }
         }
      }
      return true ;
   }
   void select(int i,int j) { 
      if (isSelect(i,j)) unselectAll() ;
      else if (isValidSelect(i,j)) setSelect(i,j) ; 
   }
   private void unselectAll() {
      for (int i=0; i<4; i++)
         for (int j=0; j<4; j++) if (isSelect(i,j)) setCounter(i,j) ;
   }
   // remove selected pieces
   void makeMove() {
      for (int i=0; i<4; i++)
         for (int j=0; j<4; j++) if (isSelect(i,j)) setEmpty(i,j) ;
   }
   // computer move
   void makeMove(Move m) {
      if (m.p1!=null) {
         setSelect(m.p1) ;
         if ((m.p2!=null)&&(isCounter(m.p2))) { 
            setSelect(m.p2) ;
            if ((m.p3!=null)&&(isCounter(m.p3))) {
               setSelect(m.p3) ;
               if ((m.p4!=null)&&(isCounter(m.p4))) setSelect(m.p4) ;
            }
         }
      }
      bc.drawBoard(this) ;
      main.sleep(500) ;
      makeMove() ;
   }
}			
/******** AI ********
   generates computer moves
*/   
final class AI {
   static final int PLAYER=0, COMPUTER=1, 
                    PLAYER_WIN=0, UNKNOWN=1, COMPUTER_WIN=2 ;
   private boolean validMove[] ;
   private Hashtable transposition = new Hashtable() ;
   private Move mv ;
   private Board bd ;
   private ControlPanel cp ;
   private TacTix main ;
   
   // constructor
   AI(Board bd,ControlPanel cp,TacTix main) { 
      this.bd = bd ; 
      this.cp = cp ; 
      this.main = main ;
   }
   // main routine
   void move() { 
      int counters = bd.getCount() ;

      if (counters==1) mv = forceMove() ;
      else {
         mv = null ;
         switch (cp.getLevel()) {
            case cp.LEVEL1: 
               if (counters>6) mv = randomMove(4) ; 
               break ;
            case cp.LEVEL2: 
               if (counters>10) mv = randomMove(3) ; 
               break ;
            case cp.LEVEL3: 
               if (counters==16) mv = randomMove(2) ;
               else if (counters>12) 
                  mv = searchLibrary(new Position(bd.getBoard())) ;
               break ;
         }
         if (mv==null) mv = searchMove() ;
      }
      bd.makeMove(mv) ;
   }
   private Move forceMove() {
      // only one counter
      for (int i=0; i<4; i++)
         for (int j=0; j<4; j++) 
            if (bd.isCounter(i,j)) 
               return new Move(new Point(i,j),null,null,null) ;
      return null ;
   }
   private Move searchLibrary(Position p) {
      // returns library move or random move if no forced-win,
      // null if position not found
      int code = p.hashCode() ;
      for (int i=0; i<Ref.library.length; i++) 
         if (code==Ref.library[i][0]) {
            if (Ref.library[i][1]>=0)
               return (Move)Ref.moveList.elementAt(Ref.library[i][1]) ;
            else return randomMove(1) ;
         }
      return null ;
   }
   private Move searchMove() {
      // returns move from alpha-beta search routine
      // or random move if no forced-win
      setValidMove() ;
      Move m = abSearch(COMPUTER,PLAYER_WIN,COMPUTER_WIN,0).getMove() ;
      if (m==null) {
         switch (cp.getLevel()) {
            case cp.LEVEL1: m = randomMove(3) ; break ;
            case cp.LEVEL2: m = randomMove(2) ; break ;
            case cp.LEVEL3: m = randomMove(1) ; break ;
         }
      }
      return m ;
   }
   private BestMove abSearch(int side,int alpha,int beta,int depth) {
      // search recursively to end of game, 
      // returning forced-win move if found, null otherwise
      int otherSide, baseCase, value ;
      Point transData=null ;
      Position thisPosition = new Position(bd.getBoard()) ;
      BestMove reply ;
      Move testMove, saveMove=null ;
      
      // base to end recursion
      baseCase = positionValue(side) ;
      if (baseCase!=UNKNOWN) return new BestMove(baseCase) ;
      
      if (depth==0) transposition.clear() ;
      // check if position in transposition table
      else if ((transData=(Point)transposition.get(thisPosition))!=null) {
         if (transData.x==side) return new BestMove(transData.y) ;
         else {
            // swap values if data is for other side
            if (transData.y==COMPUTER_WIN) return new BestMove(PLAYER_WIN) ;
            else return new BestMove(COMPUTER_WIN) ;
         }
      }
      // set initial alpha-beta values
      if (side==COMPUTER) { otherSide = PLAYER ; value = alpha ; }
      else { otherSide = COMPUTER ; value = beta ; }
      // loop for all possible moves
      for (int index=0; index<Ref.moveList.size(); index++) {
         if (!validMove[index]) continue ;
         testMove = (Move)Ref.moveList.elementAt(index) ;
         if (!tryMove(testMove)) continue ;
         makeSearchMove(testMove) ;
         // recursive search
         reply = abSearch(otherSide,alpha,beta,depth+1) ;
         unMakeSearchMove(testMove) ;
         
         if ( ((side==COMPUTER)&&(reply.getValue()>value))||
              ((side==PLAYER)&&(reply.getValue()<value)) ) {
            // save better move 
            if (side==COMPUTER) alpha = value = reply.getValue() ;
            else beta = value = reply.getValue() ;
            saveMove = testMove ;
            // refutation
            if (alpha >= beta) break ;
         }
      }
      // save board-side-value in transposition table
      transposition.put(thisPosition,new Point(side,value)) ;
      return new BestMove(value,saveMove) ;
   }
   private void setValidMove() {
      // mark all possible moves in current board position
      validMove = new boolean[64] ;
      for (int i=0; i<Ref.moveList.size(); i++)
         if (tryMove((Move)Ref.moveList.elementAt(i))) 
            validMove[i] = true ;
   }
   private int positionValue(int side) {
      if (bd.getCount()==0)
         return (side==COMPUTER)?COMPUTER_WIN:PLAYER_WIN ;
      else if (bd.getCount()==1) 
         return (side==COMPUTER)?PLAYER_WIN:COMPUTER_WIN ;
      else return UNKNOWN ;
   }
   private Move randomMove(int max) {
      // returns random move of 1 to max counters
      int rnd=0 ;
      Move rMove ;

      while(true) { 
         switch (max) {
            case 4: rnd = randomInt(Ref.moveList.size()) ; break ;
            case 3: rnd = randomInt(Ref.moveList.size()-8) + 8 ; break ;
            case 2: rnd = randomInt(Ref.moveList.size()-24) + 24 ; break ;
            case 1: rnd = randomInt(Ref.moveList.size()-48) + 48 ; break ;
         }
         rMove = (Move)Ref.moveList.elementAt(rnd) ;
         if (tryMove(rMove)) break ;
      }
      return rMove ;
   }                            
   // returns random int from 0 to n-1
   private int randomInt(int n) { return (int)(n*Math.random()) ; }

   // make-unmake routines are used in a-b search routine
   private void makeSearchMove(Move m) {
      if (m.p1!=null) {
         bd.setEmpty(m.p1) ;
         if (m.p2!=null) { 
            bd.setEmpty(m.p2) ;
            if (m.p3!=null) {
               bd.setEmpty(m.p3) ;
               if (m.p4!=null) bd.setEmpty(m.p4) ;
            }
         }
      }
   }
   private void unMakeSearchMove(Move m) {
      if (m.p1!=null) {
         bd.setCounter(m.p1) ;
         if (m.p2!=null) { 
            bd.setCounter(m.p2) ;
            if (m.p3!=null) {
               bd.setCounter(m.p3) ;
               if (m.p4!=null) bd.setCounter(m.p4) ;
            }
         }
      }
   }
   private boolean tryMove(Move m) {
      // returns true if move is possible on current board
      if (!bd.isCounter(m.p1)) return false ;
      if (m.p2!=null) { if (!bd.isCounter(m.p2)) return false ; }
      else return true ;
      if (m.p3!=null) { if (!bd.isCounter(m.p3)) return false ; }
      else return true ;
      if (m.p4!=null) { if (!bd.isCounter(m.p4)) return false ; }
      return true ;
   }
}
/******** POSITION ********
   class for representing board positions, used as key in
   transposition hashtable 
*/   
final class Position {
   private int[][] pos ;
   
   // constructors
   Position() { }
   Position(int[][] board) {
      pos = new int[4][4] ;
      for (int i=0; i<4; i++)
         for (int j=0; j<4; j++) pos[i][j] = board[i][j] ;
   }
   // hash position by converting board            16 15 14 13
   // array to 16-bit integer value.               12 11 10  9
   // bit=1 if square has counter, else 0.          8  7  6  5
   // bits are associated with squares like so:     4  3  2  1
   public int hashCode() {
      int bit=0, hashValue=0 ;
      for (int j=3; j>=0; j--)
         for (int i=3; i>=0; i--) {
            hashValue |= (pos[i][j] << bit++) ;
         }
      return hashValue ;
   }
   public boolean equals(Object otherPos) {
      for (int i=0; i<4; i++)
         for (int j=0; j<4; j++) 
            if (pos[i][j] != ((Position)otherPos).pos[i][j]) return false ;
      return true ;
   }
}
/******** MOVE ********
   represents move of 1-4 counters
*/   
final class Move {
   public Point p1, p2, p3, p4 ;

   // constructors
   Move() { }
   Move(Point p1,Point p2,Point p3,Point p4) {
      this.p1 = p1 ; this.p2 = p2; this.p3 = p3 ; this.p4 = p4;
   }
}
/******** BEST MOVE ********
   temporary storage of candidate moves, used in alpha-beta search
*/   
final class BestMove {
   private Move move ;
   private int value ;
   
   // constructors
   BestMove() { }
   BestMove(int v) { this(v, null) ; }
   BestMove(int v, Move m) {
      value = v ;
      move = m ;
   }
   Move getMove() { return move ; }
   int getValue() { return value ; }
}
/******** REF ********
   reference class, contains
      moveList - vector of all possible moves on 4x4 board
      library - array of board positions with response move
*/   
final class Ref {
   // library array has two integers values for each board position:
   // first is hashed value of board array (from Position.hashCode())
   // second is move to make, as element # from moveList vector
   //    or -1 if no winning move exists
   // library includes most opening moves, along with a few other
   // 14 and 13 counter positions with lengthy search times
   static final int[][] library = {
      // 15 counter positions - openings
      {0x7FFF,33},{0xBFFF,24},{0xDFFF,33},{0xEFFF,24},
      {0xF7FF,36},{0xFBFF,58},{0xFDFF,54},{0xFEFF,38},
      {0xFF7F,45},{0xFFBF,57},{0xFFDF,53},{0xFFEF,47},
      {0xFFF7,35},{0xFFFB,26},{0xFFFD,35},{0xFFFE,26},
      // 14 counter positions - openings
      {0x3FFF,47},{0x9FFF, 5},{0xCFFF,45},{0xF3FF,12},{0xF9FF, 4},{0xFCFF,10},
      {0xFF3F,13},{0xFF9F, 7},{0xFFCF,11},{0xFFF3,38},{0xFFF9, 6},{0xFFFC,36},
      {0x77FF,35},{0xF77F, 1},{0xFF77,33},{0xBBFF,20},{0xFBBF, 0},{0xFFBB,18},
      {0xDDFF,21},{0xFDDF, 3},{0xFFDD,19},{0xEEFF,26},{0xFEEF, 2},{0xFFEE,24},
      // 14 counter positions - forced-win
      {0x7DFF,28},{0xEBFF,31},{0xFFD7,28},{0xFFBE,31},{0x7FBF,40},{0xFBF7,43},
      {0xEFDF,40},{0xFDFE,43},{0xDF7F,28},{0xBFEF,40},{0xF7FD,43},{0xFEFB,31},
      {0xBFFB,18},{0xDFFD,19},
      // 14 counter positions - no-win
      {0xFBDF,-1},{0xFDBF,-1},{0x7FFE,-1},{0xEFF7,-1},
      {0xBFFD,-1},{0xDFFB,-1},{0xF7EF,-1},{0xFE7F,-1},
      // 13 counter positions - openings
      {0x1FFF,15},{0x8FFF, 9},{0xF1FF,17},{0xF8FF,16},
      {0xFF1F,23},{0xFF8F,22},{0xFFF1,14},{0xFFF8, 8},
      {0x777F,23},{0xF777,17},{0xBBBF, 9},{0xFBBB, 8},
      {0xDDDF,15},{0xFDDD,14},{0xEEEF,22},{0xFEEE,16},
      // 13 counter positions - forced-win
      {0x7BDF,63},{0xEDBF,51},{0xFDB7,60},{0xFBDE,48},
      {0xEBDF,57},{0xFBD7,54},{0x7DBF,53},{0xFDBE,58}
      } ;

   static final Vector moveList = new Vector(64) ;
   static {
      // vector sorted by move size, to optimize a-b search
      moveList.addElement(new Move(new Point(0,0),new Point(0,1),new Point(0,2),new Point(0,3))) ;   // 0
      moveList.addElement(new Move(new Point(1,0),new Point(1,1),new Point(1,2),new Point(1,3))) ;
      moveList.addElement(new Move(new Point(2,0),new Point(2,1),new Point(2,2),new Point(2,3))) ;   // 2
      moveList.addElement(new Move(new Point(3,0),new Point(3,1),new Point(3,2),new Point(3,3))) ;
      moveList.addElement(new Move(new Point(0,0),new Point(1,0),new Point(2,0),new Point(3,0))) ;   // 4
      moveList.addElement(new Move(new Point(0,1),new Point(1,1),new Point(2,1),new Point(3,1))) ;
      moveList.addElement(new Move(new Point(0,2),new Point(1,2),new Point(2,2),new Point(3,2))) ;   // 6
      moveList.addElement(new Move(new Point(0,3),new Point(1,3),new Point(2,3),new Point(3,3))) ;

      moveList.addElement(new Move(new Point(0,0),new Point(0,1),new Point(0,2),null)) ;   // 8
      moveList.addElement(new Move(new Point(0,1),new Point(0,2),new Point(0,3),null)) ;
      moveList.addElement(new Move(new Point(1,0),new Point(1,1),new Point(1,2),null)) ;   // 10
      moveList.addElement(new Move(new Point(1,1),new Point(1,2),new Point(1,3),null)) ;
      moveList.addElement(new Move(new Point(2,0),new Point(2,1),new Point(2,2),null)) ;   // 12
      moveList.addElement(new Move(new Point(2,1),new Point(2,2),new Point(2,3),null)) ;
      moveList.addElement(new Move(new Point(3,0),new Point(3,1),new Point(3,2),null)) ;   // 14
      moveList.addElement(new Move(new Point(3,1),new Point(3,2),new Point(3,3),null)) ;
      moveList.addElement(new Move(new Point(0,0),new Point(1,0),new Point(2,0),null)) ;   // 16
      moveList.addElement(new Move(new Point(1,0),new Point(2,0),new Point(3,0),null)) ;
      moveList.addElement(new Move(new Point(0,1),new Point(1,1),new Point(2,1),null)) ;   // 18
      moveList.addElement(new Move(new Point(1,1),new Point(2,1),new Point(3,1),null)) ;
      moveList.addElement(new Move(new Point(0,2),new Point(1,2),new Point(2,2),null)) ;   // 20
      moveList.addElement(new Move(new Point(1,2),new Point(2,2),new Point(3,2),null)) ;
      moveList.addElement(new Move(new Point(0,3),new Point(1,3),new Point(2,3),null)) ;   // 22
      moveList.addElement(new Move(new Point(1,3),new Point(2,3),new Point(3,3),null)) ;

      moveList.addElement(new Move(new Point(0,0),new Point(0,1),null,null)) ;   // 24
      moveList.addElement(new Move(new Point(0,1),new Point(0,2),null,null)) ;
      moveList.addElement(new Move(new Point(0,2),new Point(0,3),null,null)) ;   // 26
      moveList.addElement(new Move(new Point(1,0),new Point(1,1),null,null)) ;
      moveList.addElement(new Move(new Point(1,1),new Point(1,2),null,null)) ;   // 28
      moveList.addElement(new Move(new Point(1,2),new Point(1,3),null,null)) ;
      moveList.addElement(new Move(new Point(2,0),new Point(2,1),null,null)) ;   // 30
      moveList.addElement(new Move(new Point(2,1),new Point(2,2),null,null)) ;
      moveList.addElement(new Move(new Point(2,2),new Point(2,3),null,null)) ;   // 32
      moveList.addElement(new Move(new Point(3,0),new Point(3,1),null,null)) ;
      moveList.addElement(new Move(new Point(3,1),new Point(3,2),null,null)) ;   // 34
      moveList.addElement(new Move(new Point(3,2),new Point(3,3),null,null)) ;
      moveList.addElement(new Move(new Point(0,0),new Point(1,0),null,null)) ;   // 36
      moveList.addElement(new Move(new Point(1,0),new Point(2,0),null,null)) ;
      moveList.addElement(new Move(new Point(2,0),new Point(3,0),null,null)) ;   // 38
      moveList.addElement(new Move(new Point(0,1),new Point(1,1),null,null)) ;
      moveList.addElement(new Move(new Point(1,1),new Point(2,1),null,null)) ;   // 40
      moveList.addElement(new Move(new Point(2,1),new Point(3,1),null,null)) ;
      moveList.addElement(new Move(new Point(0,2),new Point(1,2),null,null)) ;   // 42
      moveList.addElement(new Move(new Point(1,2),new Point(2,2),null,null)) ;
      moveList.addElement(new Move(new Point(2,2),new Point(3,2),null,null)) ;   // 44
      moveList.addElement(new Move(new Point(0,3),new Point(1,3),null,null)) ;
      moveList.addElement(new Move(new Point(1,3),new Point(2,3),null,null)) ;   // 46
      moveList.addElement(new Move(new Point(2,3),new Point(3,3),null,null)) ;

      moveList.addElement(new Move(new Point(0,0),null,null,null)) ;   // 48
      moveList.addElement(new Move(new Point(0,1),null,null,null)) ;
      moveList.addElement(new Move(new Point(0,2),null,null,null)) ;   // 50
      moveList.addElement(new Move(new Point(0,3),null,null,null)) ;
      moveList.addElement(new Move(new Point(1,0),null,null,null)) ;   // 52
      moveList.addElement(new Move(new Point(1,1),null,null,null)) ;
      moveList.addElement(new Move(new Point(1,2),null,null,null)) ;   // 54
      moveList.addElement(new Move(new Point(1,3),null,null,null)) ;
      moveList.addElement(new Move(new Point(2,0),null,null,null)) ;   // 56
      moveList.addElement(new Move(new Point(2,1),null,null,null)) ;
      moveList.addElement(new Move(new Point(2,2),null,null,null)) ;   // 58
      moveList.addElement(new Move(new Point(2,3),null,null,null)) ;
      moveList.addElement(new Move(new Point(3,0),null,null,null)) ;   // 60
      moveList.addElement(new Move(new Point(3,1),null,null,null)) ;
      moveList.addElement(new Move(new Point(3,2),null,null,null)) ;   // 62
      moveList.addElement(new Move(new Point(3,3),null,null,null)) ;
   }
}
/******** BOARD CANVAS ********
   controls board drawing, conversion of squares/pixels
*/   
final class BoardCanvas extends Canvas {
   static final int BORDER=9, SQUARE=50, COUNTER=26, OFFSET=11 ;
   private Image boardImage, counterImage, counterOnImage, bufferImage ;
   private Graphics buffer ;
   private TacTix main ;

   // constructor
   BoardCanvas(Image bI,Image cI,Image cOnI,TacTix main) {
      boardImage = bI ;
      counterImage = cI ;
      counterOnImage = cOnI ;
      this.main = main ;
   }
	void drawBoard(Board bd) {
		if (buffer==null) {
			bufferImage = createImage(main.bcWidth,main.bcHeight) ;
			buffer = bufferImage.getGraphics() ;
		}
		// draw the board
		buffer.drawImage(boardImage,0,0,this) ;
		// add the pieces
		for (int i=0; i<4; i++) {
			for (int j=0; j<4; j++) {
				if (!bd.isEmpty(i,j)) {
					buffer.drawImage(bd.isSelect(i,j)?counterOnImage:counterImage,
						squareToPixelX(i),squareToPixelY(j),this) ;
				}
			}
		}
		repaint() ;
	}
	public void paint(Graphics g) { update(g) ; }
	public void update(Graphics g) { g.drawImage(bufferImage,0,0,this) ; }
	
   private Point pixelToSquare(int x,int y) {
      x -= BORDER ; y -= BORDER ;
      int remX=x%SQUARE, remY=y%SQUARE ;
      // if x-y is on counter, return counter position
      if ( (remX>OFFSET)&&(remX<(OFFSET+COUNTER)) &&
           (remY>OFFSET)&&(remY<(OFFSET+COUNTER)) ) {
         return new Point( (int)(x/SQUARE),(int)(y/SQUARE) ) ;
      }
      return new Point(-1,-1) ;
   }
   private int squareToPixelX(int i) { return (BORDER+(i*SQUARE)+OFFSET) ; }
   private int squareToPixelY(int j) { return (BORDER+(j*SQUARE)+OFFSET) ; }
	
	// mouse event handler
	public boolean mouseDown(Event e,int x,int y) {	
      Point p = pixelToSquare(x,y) ;
      if (p.x>=0) main.selectCounter(p.x,p.y) ; 
      return true ; 
	}
}
/******** CONTROL PANEL ********
   the UI components
*/   
final class ControlPanel extends Panel {
   static final int LEVEL1=1, LEVEL2=2, LEVEL3=3 ;
   static final Font titleFont=new Font("SansSerif", Font.BOLD, 14),
                     textFont=new Font("SansSerif", Font.PLAIN, 12),
                     statusFont=new Font("Dialog", Font.PLAIN, 12) ;
   private Label lTitle ;
   private Button bNewGame, bMove ;
   private CheckboxGroup cbgLevel ;
   private Checkbox cbBeginner, cbIntermed, cbExpert ;
   private TextField tfStatus ;
   private TacTix main ;
   
   // constructor
   ControlPanel(TacTix main) {
      this.main = main ;
      // main layout
      setLayout(new GridLayout(0,1,0,3)) ;
      setBackground(Color.lightGray) ;
      setFont(textFont) ;
      // Title
      add(lTitle=new Label(main.TITLE,Label.CENTER)) ;
      lTitle.setFont(titleFont) ;
      add(new Label(" ")) ;
      // New Game and Move buttons
      add(bNewGame=new Button(main.NEW_GAME)) ;
      // Level checkbox group
      cbgLevel=new CheckboxGroup() ;
      add(cbBeginner=new Checkbox(main.BEGINNER,cbgLevel,true)) ;
      add(cbIntermed=new Checkbox(main.INTERMED,cbgLevel,false)) ;
      add(cbExpert=new Checkbox(main.EXPERT,cbgLevel,false)) ;
      // New Game and Move buttons
      add(bMove=new Button(main.MOVE)) ;
      // Status textfield
      tfStatus = new TextField(10) ;
      tfStatus.setFont(statusFont) ;
      tfStatus.setForeground(Color.white) ;
      tfStatus.setBackground(Color.black) ;
      tfStatus.setEditable(false) ;
      add(tfStatus) ;      
      validate() ;
   }
   // handle control panel events
   public boolean action(Event e,Object o) { 
      if (main.NEW_GAME.equals(o)) main.newGame() ; 
      else if (main.MOVE.equals(o)) main.moveButton() ;
      else main.setFocus() ;
      return true ;
   }
   int getLevel() {
      if (cbBeginner.getState()) return LEVEL1 ;
      else if (cbIntermed.getState()) return LEVEL2 ;
      else if (cbExpert.getState()) return LEVEL3 ;
      return 0 ;
   }
   void setMoveButton(boolean b) { bMove.enable(b) ; }
   void setStatus(String s) { tfStatus.setText(s) ; }
}